local tetris = {}

tetris.pieces = { -- { {x1, y1}, {x2, y2}, {x3, y3}, {x4, y4} }
  { { -1,  0 }, {  0,  0 }, {  1,  0 }, {  0, -1 }, },  --  1: T up
  { {  0, -1 }, {  0,  0 }, {  1,  0 }, {  0,  1 }, },  --  2: T right
  { { -1,  0 }, {  0,  0 }, {  1,  0 }, {  0,  1 }, },  --  3: T down (spawn)
  { {  0, -1 }, { -1,  0 }, {  0,  0 }, {  0,  1 }, },  --  4: T left

  { {  0, -1 }, {  0,  0 }, { -1,  1 }, {  0,  1 }, },  --  5: J left
  { { -1, -1 }, { -1,  0 }, {  0,  0 }, {  1,  0 }, },  --  6: J up
  { {  0, -1 }, {  1, -1 }, {  0,  0 }, {  0,  1 }, },  --  7: J right
  { { -1,  0 }, {  0,  0 }, {  1,  0 }, {  1,  1 }, },  --  8: J down (spawn)

  { { -1,  0 }, {  0,  0 }, {  0,  1 }, {  1,  1 }, },  --  9: Z horizontal (spawn) 
  { {  1, -1 }, {  0,  0 }, {  1,  0 }, {  0,  1 }, },  -- 10: Z vertical

  { { -1,  0 }, {  0,  0 }, { -1,  1 }, {  0,  1 }, },  -- 11: O (spawn)

  { {  0,  0 }, {  1,  0 }, { -1,  1 }, {  0,  1 }, },  -- 12: S horizontal (spawn)
  { {  0, -1 }, {  0,  0 }, {  1,  0 }, {  1,  1 }, },  -- 13: S vertical

  { {  0, -1 }, {  0,  0 }, {  0,  1 }, {  1,  1 }, },  -- 14: L right
  { { -1,  0 }, {  0,  0 }, {  1,  0 }, { -1,  1 }, },  -- 15: L down (spawn)
  { { -1, -1 }, {  0, -1 }, {  0,  0 }, {  0,  1 }, },  -- 16: L left
  { {  1, -1 }, { -1,  0 }, {  0,  0 }, {  1,  0 }, },  -- 17: L up

  { {  0, -2 }, {  0, -1 }, {  0,  0 }, {  0,  1 }, },  -- 18: I vertical
  { { -2,  0 }, { -1,  0 }, {  0,  0 }, {  1,  0 }, }   -- 19: I horizontal (spawn)
}

tetris.spawns = {
  3, 8, 9, 11, 12, 15, 19
}

tetris.rotations = { -- {clockwise, counterclockwise}
  {  2,  4 }, --  1: T up
  {  3,  1 }, --  2: T right
  {  4,  2 }, --  3: T down
  {  1,  3 }, --  4: T left
  
  {  6,  8 }, --  5: J left
  {  7,  5 }, --  6: J up
  {  8,  6 }, --  7: J right
  {  5,  7 }, --  8: J down
  
  { 10, 10 }, --  9: Z horizontal
  {  9,  9 }, -- 10: Z vertical
  
  { 11, 11 }, -- 11: O
  
  { 13, 13 }, -- 12: S horizontal
  { 12, 12 }, -- 13: S vetical
  
  { 15, 17 }, -- 14: L right
  { 16, 14 }, -- 15: L down
  { 17, 15 }, -- 16: L left
  { 14, 16 }, -- 17: L up
  
  { 19, 19 }, -- 18: I vertical
  { 18, 18 }  -- 19: I horizontal
}

tetris.tiletypes = {
  3, 3, 3, 3, 1, 1, 1, 1, 2, 2, 3, 1, 1, 2, 2, 2, 2, 3, 3
}

tetris.framerate = 60.0988

tetris.fallspeed = {
  6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1
}

tetris.input = {
  dir = 0,
  das = 0,
  rot = 0
}

tetris.das = {
  delay = 16,
  ["repeat"] = 6
}

function tetris:newPiece(id)
  return {
    x = 6,
    y = 1,
    id = id
  }
end

function tetris:newPieceId()
  local id = self.spawns[math.random(#self.spawns)]
  return id
end

function tetris:newGame(level)
  self.game = {
    level = level or 0,
    lines = 0,
    score = 0,
    levellines = 0,
    piece = self:newPiece(self:newPieceId()),
    nextpiece = self:newPieceId(),
    fallcountdown = 0,
    board = {
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
    },
    lineclearqueue = {}, -- template: { line = [number], state = [number, 0 - 5] }
    topout = false
  }
end

function tetris:fall()
  --if not playing a game, having no current piece, topped out or clearing lines, do nothing
  if not self.game or not self.game.piece or self.game.topout or #self.game.lineclearqueue > 0 then return end
  
  local piece,board = self.pieces[self.game.piece.id], self.game.board
  --if the piece is on ground, place all its (inbound) tiles into the board
  if self:pieceCheck(self.game.piece.x, self.game.piece.y + 1, self.game.piece.id) then
    for i=1,4 do
      local x,y = self.game.piece.x + piece[i][1], self.game.piece.y + piece[i][2]
      if x > 0 and x < 11 and y > 0 and y < 21 then
        self.game.board[y][x] = self.tiletypes[self.game.piece.id]
      end
    end
  end
  --if can't place piece, topout
  if self:pieceCheck(6,  1, self.game.nextpiece) then
    self.game.topout = true
    print("topped out on level " .. tostring(self.game.level) .. " with " .. tostring(self.game.lines) .. " lines and a score of " .. tostring(self.game.score))
    print("board on topout: ")
    for y=1,20 do
      local str = ""
      for x=1,10 do
        str = str .. self.game.board[y][x] .. " "
      end
      print(str)
    end
  end
  --move the piece from the queue into the playfield and generate the next piece
  self.game.piece = self:newPiece(self.game.nextpiece)
  self.game.nextpiece = self:newPieceId()
end

function tetris:update(dt, dir, rot)
  --if not playing a game, having no current piece, topped out or clearing lines, do nothing
  if not self.game or not self.game.piece or self.game.topout or #self.game.lineclearqueue > 0 then return 0 end
  
  local frames = math.ceil(dt * self.framerate)
  local wait = (frames / self.framerate) - dt
  
  local board = self.game.board
  --update the clear queue
  local clearqueue = self.game.lineclearqueue
  self.game.lineclearqueue = {}
  for i,lc in ipairs(clearqueue) do
    --increment the state
    lc.state = lc.state + 1
    if lc.state >= 0 and lc.state <= 23 then
      --if the state has not yet passed the limit, reinsert it into the queue
      table.insert(self.game.lineclearqueue)
      --update the line
      local line = self.game.board[lc.line]
      for i=0,lc.state do
        if math.floor(i / 4) > 0 then
          line[5 + math.floor(i / 4)], line[6 - math.floor(i / 2)] = 0,0
        end
      end
      self.game.board[lc.line] = line
    else
      --if past the limit, delete the line from the board and dont reinsert the state
      table.remove(self.game.board, lc.line)
      table.insert(self.game.board, 1, {0, 0, 0, 0, 0, 0, 0, 0, 0, 0})
    end
  end
  --add full lines to the clear queue
  local clearedlines = 0
  for i=1,20 do
    local full = true
    for j=1,10 do
      if board[i][j] < 1 then
        full = false
      end
    end
    if full then
      table.insert(self.game.lineclearqueue, {line = i, state = 0})
      clearedlines = clearedlines + 1
      print("clear line " .. tonumber(i))
    end
  end
  local sm = 0
  if clearedlines == 1 then
    --single
    sm = 40
  elseif clearedlines == 2 then
    --double
    sm = 100
  elseif clearedlines == 3 then
    --triple
    sm = 300
  elseif clearedlines >= 4 then
    --tetris
    sm = 1200
  end
  self.game.lines = self.game.lines + clearedlines
  self.game.levellines = self.game.levellines + clearedlines
  if self.game.levellines > (self.game.level + 1) * 10 then
    self.game.levellines = self.game.levellines - (self.game.level + 1) * 10
    self.game.level = self.game.level + 1
  elseif self.game.levellines > 100 and self.game.levellines > (self.game.level - 5) * 10 then
    if (self.game.level - 5) * 10 > 100 then
      self.game.levellines = self.game.levellines - (self.game.level - 5) * 10
    else
      self.game.levellines = self.game.levellines - 100
    end
    self.game.level = self.game.level + 1
  end
  self.game.score = self.game.score + sm * (self.game.level + 1)
  --if clearing any lines, don't make the piece fall
  if #clearqueue > 0 then return wait end
  
  local speedindex = self.game.level
  if speedindex > #self.fallspeed then speedindex = #self.fallspeed end
  self.game.fallcountdown = self.game.fallcountdown + frames
  while self.game.fallcountdown > self.fallspeed[speedindex] do
    self:fall()
    self.game.fallcountdown = self.game.fallcountdown - self.fallspeed[speedindex]
  end
  
  self:updateInput(frames, dir or 0, rot or 0)
  
  return wait
end

function tetris:updateInput(frames, dir, rot)
  if self.input.dir == 0 and dir ~= 0 then
    --tap
    self.input.das = 0
  elseif self.input.dir ~= 0 and dir ~= 0 then
    --held
    self.input.das = self.input.das + frames
    if self.input.das > self.das.delay then
      local m = self:movePiece(dir)
      if m then
        self.input.das = self.input.das - self.das["repeat"]
      else
        self.input.das = self.das.delay
      end
    end
  end
  self.input.dir = dir
  
  if self.input.rot == 0 and rot ~= 0 then
    self:rotatePiece(rot)
  end
end

function tetris:pieceCheck(x, y, id)
  local piece = self.pieces[id]
  local col = false
  for i=1,4 do
    local tx, ty = x + piece[i][1], y + piece[i][2]
    col = col or (tx < 1 or tx > 10 or ty > 20)
    if tx > 0 and tx < 11 and ty > 0 and ty < 21 then col = col or (self.game.board[ty][tx] > 0) end
  end
  if col then 
    local piecestr = "{ "
    for i=1,4 do
      piecestr = piecestr .. "{" .. tostring(piece[i][1]) .. "," .. tostring(piece[i][2] .. "}") .. " "
      if i == 4 then piecestr = piecestr .. "}" end
    end
    print("piece " .. tostring(id) .. " collision at x=" .. tostring(x) .. ",y=" .. tostring(y) .. ", piece structure: " .. piecestr)
  end
  return col
end

function tetris:movePiece(dir)
  if not self.game or not self.game.piece or self.game.topout or #self.game.lineclearqueue > 0 then return end
  
  local canMove = not tetris:pieceCheck(self.game.piece.x + dir, self.game.piece.y, self.game.piece.id)
  if canMove then
    self.game.piece.x = self.game.piece.x + dir
  end
  
  return canMove
end

function tetris:rotatePiece(rot)
  if not self.game or not self.game.piece or self.game.topout or #self.game.lineclearqueue > 0 then return end
  
  local canRotate = false
  if rot > 0 then
    canRotate = self:pieceCheck(self.game.piece.x, self.game.piece.y, self.rotations[self.game.piece.id][1])
    if canRotate then
      self.game.piece.id = self.rotations[self.game.piece.id][1]
    end
  elseif rot < 0 then
    canRotate = self:pieceCheck(self.game.piece.x, self.game.piece.y, self.rotations[self.game.piece.id][2])
    if canRotate then
      self.game.piece.id = self.rotations[self.game.piece.id][2]
    end
  end
  
  return canRotate
end

function tetris:setDasDelay(d)
  self.das.delay = d
end

function tetris:setDasRepeat(r)
  self.das["repeat"] = r
end

function tetris:setDas(d, r)
  self.das = {delay = d, ["repeat"] = r}
end

function tetris:getDas()
  return self.input.das
end

function tetris:getStats()
  return self.game and {
    level = self.game.level or 0,
    lines = self.game.lines or 0,
    score = self.game.score or 0,
    topout = self.game and self.game.topout
  }
end

function tetris:getBoard(x, y)
  return self.game and self.game.board and self.game.board[y][x]
end

function tetris:getPlayfield()
  if self.game and self.game.board and self.game.piece then
    local board = self.game.board
    for i=1,4 do
      local x,y = self.game.piece.x + self.pieces[self.game.piece.id][i][1], self.game.piece.y + self.pieces[self.game.piece.id][i][2]
      if x > 0 and x < 11 and y > 0 and y < 21 then
        board[y][x] = self.tiletypes[self.game.piece.id]
      end
    end
    
    return board
  end
end

function tetris:getNextPiece()
  return self.game.nextpiece and {
    tiles = self.pieces[self.game.nextpiece],
    tiletype = self.tiletypes[self.game.nextpiece]
  }
end

function tetris:endGame()
  self.game = false
end

return tetris